//gcc -g -o libpthreaddbg.so -shared pthreaddbg.c -ldl

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <pthread.h>
#include <gnu/lib-names.h>
#include <execinfo.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

//The following represents how much of the stack is already used
//When we can 1st access it.  It is likely to be different for
//different pthread libs and platforms
#define PTHREAD_OFFSET_MAGIC 0x1ba8
static int fd;
static char printstr[2048];
static int threadcount = 0;
static pthread_mutex_t printlock = PTHREAD_MUTEX_INITIALIZER;

struct pstack_t {
  int    used;
  int    thread_num;
  unsigned int start;
  unsigned int end;
};
#define PSTACK_SIZE 4096
struct pstack_t *pstack;

static int (*ppthread_create)(pthread_t *__restrict __threadp,
			   __const pthread_attr_t *__restrict __attr,
			   void *(*__start_routine) (void *),
			   void *__restrict __arg) ;
static pthread_t (*ppthread_self)(void) ;
static int (*ppthread_equal)(pthread_t __thread1, pthread_t __thread2) ;
static void (*ppthread_exit)(void *__retval)  __attribute__ ((__noreturn__));
static int (*ppthread_join)(pthread_t __th, void **__thread_return);
static int (*ppthread_detach)(pthread_t __th) ;
static int (*ppthread_attr_init)(pthread_attr_t *__attr) ;
static int (*ppthread_attr_destroy)(pthread_attr_t *__attr) ;
static int (*ppthread_attr_setdetachstate)(pthread_attr_t *__attr,
					int __detachstate) ;
static int (*ppthread_attr_getdetachstate)(__const pthread_attr_t *__attr,
					int *__detachstate) ;
static int (*ppthread_attr_setschedparam)(pthread_attr_t *__restrict __attr,
				       __const struct sched_param *__restrict
				       __param) ;
static int (*ppthread_attr_getschedparam)(__const pthread_attr_t *__restrict
				       __attr,
				       struct sched_param *__restrict __param)
     ;
static int (*ppthread_attr_setschedpolicy)(pthread_attr_t *__attr, int __policy)
     ;
static int (*ppthread_attr_getschedpolicy)(__const pthread_attr_t *__restrict
					__attr, int *__restrict __policy)
     ;
static int (*ppthread_attr_setinheritsched)(pthread_attr_t *__attr,
					 int __inherit) ;
static int (*ppthread_attr_getinheritsched)(__const pthread_attr_t *__restrict
					 __attr, int *__restrict __inherit)
     ;
static int (*ppthread_attr_setscope)(pthread_attr_t *__attr, int __scope)
     ;
static int (*ppthread_attr_getscope)(__const pthread_attr_t *__restrict __attr,
				  int *__restrict __scope) ;
static int (*ppthread_attr_setstackaddr)(pthread_attr_t *__attr,
				      void *__stackaddr) ;
static int (*ppthread_attr_getstackaddr)(__const pthread_attr_t *__restrict
				      __attr, void **__restrict __stackaddr)
     ;
static int (*ppthread_attr_setstacksize)(pthread_attr_t *__attr,
				      size_t __stacksize) ;
static int (*ppthread_attr_getstacksize)(__const pthread_attr_t *__restrict
				      __attr, size_t *__restrict __stacksize)
     ;
static int (*ppthread_getattr_np)(pthread_t __th, pthread_attr_t *__attr) ;
static int (*ppthread_setschedparam)(pthread_t __target_thread, int __policy,
				  __const struct sched_param *__param)
     ;
static int (*ppthread_getschedparam)(pthread_t __target_thread,
				  int *__restrict __policy,
				  struct sched_param *__restrict __param)
     ;
static int (*ppthread_yield)(void) ;
static int (*ppthread_mutex_init)(pthread_mutex_t *__restrict __mutex,
			       __const pthread_mutexattr_t *__restrict
			       __mutex_attr) ;
static int (*ppthread_mutex_destroy)(pthread_mutex_t *__mutex) ;
static int (*ppthread_mutex_trylock)(pthread_mutex_t *__mutex) ;
static int (*ppthread_mutex_lock)(pthread_mutex_t *__mutex) ;
static int (*ppthread_mutex_unlock)(pthread_mutex_t *__mutex) ;
static int (*ppthread_mutexattr_init)(pthread_mutexattr_t *__attr) ;
static int (*ppthread_mutexattr_destroy)(pthread_mutexattr_t *__attr) ;
static int (*ppthread_mutexattr_getpshared)(__const pthread_mutexattr_t *
					 __restrict __attr,
					 int *__restrict __pshared) ;
static int (*ppthread_mutexattr_setpshared)(pthread_mutexattr_t *__attr,
					 int __pshared) ;
static int (*ppthread_cond_init)(pthread_cond_t *__restrict __cond,
			      __const pthread_condattr_t *__restrict
			      __cond_attr) ;
static int (*ppthread_cond_destroy)(pthread_cond_t *__cond) ;
static int (*ppthread_cond_signal)(pthread_cond_t *__cond) ;
static int (*ppthread_cond_broadcast)(pthread_cond_t *__cond) ;
static int (*ppthread_cond_wait)(pthread_cond_t *__restrict __cond,
			      pthread_mutex_t *__restrict __mutex);
static int (*ppthread_cond_timedwait)(pthread_cond_t *__restrict __cond,
				   pthread_mutex_t *__restrict __mutex,
				   __const struct timespec *__restrict
				   __abstime);
static int (*ppthread_condattr_init)(pthread_condattr_t *__attr) ;
static int (*ppthread_condattr_destroy)(pthread_condattr_t *__attr) ;
static int (*ppthread_condattr_getpshared)(__const pthread_condattr_t *
					__restrict __attr,
					int *__restrict __pshared) ;
static int (*ppthread_condattr_setpshared)(pthread_condattr_t *__attr,
					int __pshared) ;
static int (*ppthread_key_create)(pthread_key_t *__key,
			       void (*__destr_function) (void *)) ;
static int (*ppthread_key_delete)(pthread_key_t __key) ;
static int (*ppthread_setspecific)(pthread_key_t __key,
				__const void *__pointer) ;
static void *(*ppthread_getspecific)(pthread_key_t __key) ;

static int (*ppthread_once)(pthread_once_t *__once_control,
			 void (*__init_routine) (void));
static int (*ppthread_setcancelstate)(int __state, int *__oldstate);
static int (*ppthread_setcanceltype)(int __type, int *__oldtype);
static int (*ppthread_cancel)(pthread_t __cancelthread);
static void (*ppthread_testcancel)(void);
static void (*_ppthread_cleanup_push)(struct _pthread_cleanup_buffer *__buffer,
				   void (*__routine) (void *),
				   void *__arg) ;
static void (*_ppthread_cleanup_pop)(struct _pthread_cleanup_buffer *__buffer,
				  int __execute) ;
static void (*_ppthread_cleanup_push_defer)(struct _pthread_cleanup_buffer *__buffer,
					 void (*__routine) (void *),
					 void *__arg) ;
static void (*_ppthread_cleanup_pop_restore)(struct _pthread_cleanup_buffer *__buffer,
					  int __execute) ;
static int (*ppthread_atfork)(void (*__prepare) (void),
			   void (*__parent) (void),
			   void (*__child) (void)) ;
static void (*ppthread_kill_other_threads_np)(void) ;
static int (*ppthread_kill) (pthread_t thread, int sig);

/***************************************************/
typedef struct {
  void *(*cmd) (void *);
  void *arg;
  void *array[10];
  size_t arraysize;
} threadinfo;

void finish(int loc)
{
  ppthread_mutex_lock(&printlock);
  if(loc < 0) {
    register void *ebp __asm__ ("ebp");
    unsigned int stack = (unsigned int)(ebp);
    for(loc = 0; loc < PSTACK_SIZE; loc++) {
      if(pstack[loc].used && stack > pstack[loc].start &&
                             stack < pstack[loc].end) {
         break;
      }
    }
  }
  if(loc < PSTACK_SIZE) {
    fprintf(stderr, "Terminating thread #%d\n", pstack[loc].thread_num);
    pstack[loc].used = 0;
  }
  ppthread_mutex_unlock(&printlock);
}

static void *thread_stub(void *data) {
  register void *ebp __asm__ ("ebp");
  pthread_attr_t attr;
  unsigned int stack;
  size_t size;
  char **strings;
  char buf[80];
  time_t t;
  int loc;
  void *ret;
  threadinfo ti = *((threadinfo *)data);
  free(data);
  stack = (unsigned int)(ebp) + PTHREAD_OFFSET_MAGIC;
  ppthread_attr_init(&attr);
  ppthread_attr_getstacksize(&attr, &size);
  strings = backtrace_symbols (ti.array, ti.arraysize);
  time(&t);
  ctime_r(&t, buf);
  ppthread_mutex_lock(&printlock);
  sprintf(printstr, "\nThread #%d: %s%08x: Stack addr=%08x - %08x, size=%08x\n", threadcount++, buf,
          ppthread_self(), stack-size, stack, size);
  sprintf(printstr+strlen(printstr), "Obtained %zd stack frames.\n", ti.arraysize);
  for(loc=0; loc < PSTACK_SIZE; loc++) {
    if(! pstack[loc].used) {
       pstack[loc].used = 1;
       pstack[loc].thread_num = threadcount-1;
       pstack[loc].start = stack-size;
       pstack[loc].end = stack;
       break;
    }
  }
  for (stack = 0; stack < ti.arraysize; stack++) {
    Dl_info dli;
    dladdr(ti.array[stack], &dli);
    //fprintf(stderr, "  name:%s\n  base:%s\n  sname:%s\n  addr: %08x(%08x)\n",
    //  dli.dli_fname, dli.dli_fbase, dli.dli_sname, dli.dli_saddr, ti.array[stack]);
    sprintf(printstr+strlen(printstr), "%s\n", strings[stack]);
  }
  fprintf(stderr, printstr);
  //write(fd, printstr, strlen(printstr));
  //fsync(fd);
  ppthread_mutex_unlock(&printlock);
  free (strings);
  ret = ti.cmd(ti.arg);
  finish(loc);
  return(ret);
}
extern int pthread_create (pthread_t *__restrict __threadp,
			   __const pthread_attr_t *__restrict __attr,
			   void *(*__start_routine) (void *),
			   void *__restrict __arg)  {
  threadinfo *ti = malloc(sizeof(threadinfo));
  ti->cmd = __start_routine;
  ti->arg = __arg;
  ti->arraysize = backtrace(ti->array, 10);
  //return ppthread_create(__threadp, __attr, __start_routine, __arg);
  return ppthread_create(__threadp, __attr, thread_stub, ti);
}
extern pthread_t pthread_self (void)  {
  return ppthread_self();
}

extern int pthread_equal (pthread_t __thread1, pthread_t __thread2)  {
  return ppthread_equal(__thread1, __thread2);
}

extern void pthread_exit (void *__retval) {
  finish(-1);
  ppthread_exit(__retval);
}

extern int pthread_join (pthread_t __th, void **__thread_return) {
  return ppthread_join(__th, __thread_return);
}

extern int pthread_detach (pthread_t __th)  {
  return ppthread_detach(__th);
}

extern int pthread_attr_init (pthread_attr_t *__attr)  {
  return ppthread_attr_init(__attr);
}

extern int pthread_attr_destroy (pthread_attr_t *__attr)  {
  return ppthread_attr_destroy(__attr);
}

extern int pthread_attr_setdetachstate (pthread_attr_t *__attr,
					int __detachstate)  {
  return ppthread_attr_setdetachstate(__attr, __detachstate);
}

extern int pthread_attr_getdetachstate (__const pthread_attr_t *__attr,
					int *__detachstate)  {
  return ppthread_attr_getdetachstate(__attr, __detachstate);
}

extern int pthread_attr_setschedparam (pthread_attr_t *__restrict __attr,
				       __const struct sched_param *__restrict
				       __param)  {
  return ppthread_attr_setschedparam(__attr, __param);
}

extern int pthread_attr_getschedparam (__const pthread_attr_t *__restrict
				       __attr,
				       struct sched_param *__restrict __param)
      {
  return ppthread_attr_getschedparam(__attr, __param);
}

extern int pthread_attr_setschedpolicy (pthread_attr_t *__attr, int __policy)
      {
  return ppthread_attr_setschedpolicy(__attr, __policy);
}

extern int pthread_attr_getschedpolicy (__const pthread_attr_t *__restrict
					__attr, int *__restrict __policy)
      {
  return ppthread_attr_getschedpolicy(__attr, __policy);
}

extern int pthread_attr_setinheritsched (pthread_attr_t *__attr,
					 int __inherit)  {
  return ppthread_attr_setinheritsched(__attr, __inherit);
}

extern int pthread_attr_getinheritsched (__const pthread_attr_t *__restrict
					 __attr, int *__restrict __inherit)
      {
  return ppthread_attr_getinheritsched(__attr, __inherit);
}

extern int pthread_attr_setscope (pthread_attr_t *__attr, int __scope)
      {
  return ppthread_attr_setscope(__attr, __scope);
}

extern int pthread_attr_getscope (__const pthread_attr_t *__restrict __attr,
				  int *__restrict __scope)  {
  return ppthread_attr_getscope(__attr, __scope);
}

extern int pthread_attr_setstackaddr (pthread_attr_t *__attr,
				      void *__stackaddr)  {
  return ppthread_attr_setstackaddr(__attr, __stackaddr);
}

extern int pthread_attr_getstackaddr (__const pthread_attr_t *__restrict
				      __attr, void **__restrict __stackaddr)
      {
  return ppthread_attr_getstackaddr(__attr, __stackaddr);
}

extern int pthread_attr_setstacksize (pthread_attr_t *__attr,
				      size_t __stacksize)  {
  return ppthread_attr_setstacksize(__attr, __stacksize);
}

extern int pthread_attr_getstacksize (__const pthread_attr_t *__restrict
				      __attr, size_t *__restrict __stacksize)
      {
  return ppthread_attr_getstacksize(__attr, __stacksize);
}

extern int pthread_getattr_np (pthread_t __th, pthread_attr_t *__attr)  {
  return ppthread_getattr_np(__th, __attr);
}

extern int pthread_setschedparam (pthread_t __target_thread, int __policy,
				  __const struct sched_param *__param)
      {
  return ppthread_setschedparam(__target_thread, __policy, __param);
}

extern int pthread_getschedparam (pthread_t __target_thread,
				  int *__restrict __policy,
				  struct sched_param *__restrict __param)
      {
  return ppthread_getschedparam(__target_thread, __policy, __param);
}

extern int pthread_yield (void)  {
  return ppthread_yield();
}

extern int pthread_mutex_init (pthread_mutex_t *__restrict __mutex,
			       __const pthread_mutexattr_t *__restrict
			       __mutex_attr)  {
  return ppthread_mutex_init(__mutex, __mutex_attr);
}

extern int pthread_mutex_destroy (pthread_mutex_t *__mutex)  {
  return ppthread_mutex_destroy(__mutex);
}

extern int pthread_mutex_trylock (pthread_mutex_t *__mutex)  {
  return ppthread_mutex_trylock(__mutex);
}

extern int pthread_mutex_lock (pthread_mutex_t *__mutex)  {
  return ppthread_mutex_lock(__mutex);
}

extern int pthread_mutex_unlock (pthread_mutex_t *__mutex)  {
  return ppthread_mutex_unlock(__mutex);
}

extern int pthread_mutexattr_init (pthread_mutexattr_t *__attr)  {
  return ppthread_mutexattr_init(__attr);
}

extern int pthread_mutexattr_destroy (pthread_mutexattr_t *__attr)  {
  return ppthread_mutexattr_destroy(__attr);
}

extern int pthread_mutexattr_getpshared (__const pthread_mutexattr_t *
					 __restrict __attr,
					 int *__restrict __pshared)  {
  return ppthread_mutexattr_getpshared(__attr, __pshared);
}

extern int pthread_mutexattr_setpshared (pthread_mutexattr_t *__attr,
					 int __pshared)  {
  return ppthread_mutexattr_setpshared(__attr, __pshared);
}

extern int pthread_cond_init (pthread_cond_t *__restrict __cond,
			      __const pthread_condattr_t *__restrict
			      __cond_attr)  {
  return ppthread_cond_init(__cond, __cond_attr);
}

extern int pthread_cond_destroy (pthread_cond_t *__cond)  {
  return ppthread_cond_destroy(__cond);
}

extern int pthread_cond_signal (pthread_cond_t *__cond)  {
  return ppthread_cond_signal(__cond);
}

extern int pthread_cond_broadcast (pthread_cond_t *__cond)  {
  return ppthread_cond_broadcast(__cond);
}

extern int pthread_cond_wait (pthread_cond_t *__restrict __cond,
			      pthread_mutex_t *__restrict __mutex) {
  return ppthread_cond_wait(__cond, __mutex);
}

extern int pthread_cond_timedwait (pthread_cond_t *__restrict __cond,
				   pthread_mutex_t *__restrict __mutex,
				   __const struct timespec *__restrict
				   __abstime) {
  return ppthread_cond_timedwait(__cond, __mutex, __abstime);
}

extern int pthread_condattr_init (pthread_condattr_t *__attr)  {
  return ppthread_condattr_init(__attr);
}

extern int pthread_condattr_destroy (pthread_condattr_t *__attr)  {
  return ppthread_condattr_destroy(__attr);
}

extern int pthread_condattr_getpshared (__const pthread_condattr_t *
					__restrict __attr,
					int *__restrict __pshared)  {
  return ppthread_condattr_getpshared(__attr, __pshared);
}

extern int pthread_condattr_setpshared (pthread_condattr_t *__attr,
					int __pshared)  {
  return ppthread_condattr_setpshared(__attr, __pshared);
}

extern int pthread_key_create (pthread_key_t *__key,
			       void (*__destr_function) (void *))  {
  return ppthread_key_create(__key, __destr_function);
}

extern int pthread_key_delete (pthread_key_t __key)  {
  return ppthread_key_delete(__key);
}

extern int pthread_setspecific (pthread_key_t __key,
				__const void *__pointer)  {
  return ppthread_setspecific(__key, __pointer);
}

extern void *pthread_getspecific (pthread_key_t __key)  {
  return ppthread_getspecific(__key);
}

extern int pthread_once (pthread_once_t *__once_control,
			 void (*__init_routine) (void)) {
  return ppthread_once(__once_control, __init_routine);
}

extern int pthread_setcancelstate (int __state, int *__oldstate) {
  return ppthread_setcancelstate(__state, __oldstate);
}

extern int pthread_setcanceltype (int __type, int *__oldtype) {
  return ppthread_setcanceltype(__type, __oldtype);
}

extern int pthread_cancel (pthread_t __cancelthread) {
  return ppthread_cancel(__cancelthread);
}

extern void pthread_testcancel (void) {
  ppthread_testcancel();
}

extern void _pthread_cleanup_push (struct _pthread_cleanup_buffer *__buffer,
				   void (*__routine) (void *),
				   void *__arg)  {
  _ppthread_cleanup_push(__buffer, __routine, __arg);
}

extern void _pthread_cleanup_pop (struct _pthread_cleanup_buffer *__buffer,
				  int __execute)  {
  _ppthread_cleanup_pop(__buffer, __execute);
}

extern void _pthread_cleanup_push_defer (struct _pthread_cleanup_buffer *__buffer,
					 void (*__routine) (void *),
					 void *__arg)  {
  _ppthread_cleanup_push_defer(__buffer, __routine, __arg);
}

extern void _pthread_cleanup_pop_restore (struct _pthread_cleanup_buffer *__buffer,
					  int __execute)  {
  _ppthread_cleanup_pop_restore(__buffer, __execute);
}

extern int pthread_atfork (void (*__prepare) (void),
			   void (*__parent) (void),
			   void (*__child) (void))  {
  return ppthread_atfork(__prepare, __parent, __child);
}

extern void pthread_kill_other_threads_np (void)  {
  ppthread_kill_other_threads_np();
}

extern int pthread_kill(pthread_t thread, int sig) {
  return ppthread_kill(thread, sig);
}
int __attribute__((constructor)) __init(void)
{
  void *handle;
  handle = dlopen(LIBPTHREAD_SO, RTLD_LAZY | RTLD_LOCAL);
  if(!handle) {
    fputs("Failed:\n", stderr);
    fputs(dlerror(), stderr);
    fputs("\n", stderr);
    exit(1);
  }
  ppthread_create = dlsym(handle,"pthread_create");
  ppthread_self = dlsym(handle,"pthread_self");
  ppthread_equal = dlsym(handle,"pthread_equal");
  ppthread_exit = dlsym(handle,"pthread_exit");
  ppthread_join = dlsym(handle,"pthread_join");
  ppthread_detach = dlsym(handle,"pthread_detach");
  ppthread_attr_init = dlsym(handle,"pthread_attr_init");
  ppthread_attr_destroy = dlsym(handle,"pthread_attr_destroy");
  ppthread_attr_setdetachstate = dlsym(handle,"pthread_attr_setdetachstate");
  ppthread_attr_getdetachstate = dlsym(handle,"pthread_attr_getdetachstate");
  ppthread_attr_setschedparam = dlsym(handle,"pthread_attr_setschedparam");
  ppthread_attr_getschedparam = dlsym(handle,"pthread_attr_getschedparam");
  ppthread_attr_setschedpolicy = dlsym(handle,"pthread_attr_setschedpolicy");
  ppthread_attr_getschedpolicy = dlsym(handle,"pthread_attr_getschedpolicy");
  ppthread_attr_setinheritsched = dlsym(handle,"pthread_attr_setinheritsched");
  ppthread_attr_getinheritsched = dlsym(handle,"pthread_attr_getinheritsched");
  ppthread_attr_setscope = dlsym(handle,"pthread_attr_setscope");
  ppthread_attr_getscope = dlsym(handle,"pthread_attr_getscope");
  ppthread_attr_setstackaddr = dlsym(handle,"pthread_attr_setstackaddr");
  ppthread_attr_getstackaddr = dlsym(handle,"pthread_attr_getstackaddr");
  ppthread_attr_setstacksize = dlsym(handle,"pthread_attr_setstacksize");
  ppthread_attr_getstacksize = dlsym(handle,"pthread_attr_getstacksize");
  ppthread_getattr_np = dlsym(handle,"pthread_getattr_np");
  ppthread_setschedparam = dlsym(handle,"pthread_setschedparam");
  ppthread_getschedparam = dlsym(handle,"pthread_getschedparam");
  ppthread_yield = dlsym(handle,"pthread_yield");
  ppthread_mutex_init = dlsym(handle,"pthread_mutex_init");
  ppthread_mutex_destroy = dlsym(handle,"pthread_mutex_destroy");
  ppthread_mutex_trylock = dlsym(handle,"pthread_mutex_trylock");
  ppthread_mutex_lock = dlsym(handle,"pthread_mutex_lock");
  ppthread_mutex_unlock = dlsym(handle,"pthread_mutex_unlock");
  ppthread_mutexattr_init = dlsym(handle,"pthread_mutexattr_init");
  ppthread_mutexattr_destroy = dlsym(handle,"pthread_mutexattr_destroy");
  ppthread_mutexattr_getpshared = dlsym(handle,"pthread_mutexattr_getpshared");
  ppthread_mutexattr_setpshared = dlsym(handle,"pthread_mutexattr_setpshared");
  ppthread_cond_init = dlsym(handle,"pthread_cond_init");
  ppthread_cond_destroy = dlsym(handle,"pthread_cond_destroy");
  ppthread_cond_signal = dlsym(handle,"pthread_cond_signal");
  ppthread_cond_broadcast = dlsym(handle,"pthread_cond_broadcast");
  ppthread_cond_wait = dlsym(handle,"pthread_cond_wait");
  ppthread_cond_timedwait = dlsym(handle,"pthread_cond_timedwait");
  ppthread_condattr_init = dlsym(handle,"pthread_condattr_init");
  ppthread_condattr_destroy = dlsym(handle,"pthread_condattr_destroy");
  ppthread_condattr_getpshared = dlsym(handle,"pthread_condattr_getpshared");
  ppthread_condattr_setpshared = dlsym(handle,"pthread_condattr_setpshared");
  ppthread_key_create = dlsym(handle,"pthread_key_create");
  ppthread_key_delete = dlsym(handle,"pthread_key_delete");
  ppthread_setspecific = dlsym(handle,"pthread_setspecific");
  ppthread_getspecific = dlsym(handle,"pthread_getspecific");
  ppthread_once = dlsym(handle,"pthread_once");
  ppthread_setcancelstate = dlsym(handle,"pthread_setcancelstate");
  ppthread_setcanceltype = dlsym(handle,"pthread_setcanceltype");
  ppthread_cancel = dlsym(handle,"pthread_cancel");
  ppthread_testcancel = dlsym(handle,"pthread_testcancel");
  _ppthread_cleanup_push = dlsym(handle,"_pthread_cleanup_push");
  _ppthread_cleanup_pop = dlsym(handle,"_pthread_cleanup_pop");
  _ppthread_cleanup_push_defer = dlsym(handle,"_pthread_cleanup_push_defer");
  _ppthread_cleanup_pop_restore = dlsym(handle,"_pthread_cleanup_pop_restore");
  ppthread_atfork = dlsym(handle,"pthread_atfork");
  ppthread_kill_other_threads_np = dlsym(handle,"pthread_kill_other_threads_np");
  ppthread_kill = dlsym(handle, "pthread_kill");
  if(0) {
    int i = getpid();
    char filename[80];
    char str[80];
    sprintf(filename, "/tmp/pthreaddbg.%d", i);
    fd = open(filename, O_CREAT | O_TRUNC, 0666);
    if(fd == -1) { exit(-1);}
    sprintf(str, "Beginning pthread debug\n");
    write(fd, str, strlen(str));
  }
  pstack = (struct pstack_t *)calloc(PSTACK_SIZE,sizeof(struct pstack_t));
  return 0;
}
